from __future__ import annotations

import types
from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import nullcontext
from typing import TYPE_CHECKING, cast

import numpy as np
from numpy import typing as npt

from napari.layers import Layer
from napari.layers._data_protocols import LayerDataProtocol
from napari.layers._multiscale_data import MultiScaleData
from napari.layers.image._image_constants import Interpolation, VolumeDepiction
from napari.layers.image._image_mouse_bindings import (
    move_plane_along_normal as plane_drag_callback,
    set_plane_position as plane_double_click_callback,
)
from napari.layers.image._image_utils import guess_multiscale
from napari.layers.image._slice import _ImageSliceRequest, _ImageSliceResponse
from napari.layers.utils._slice_input import _SliceInput, _ThickNDSlice
from napari.layers.utils.plane import SlicingPlane
from napari.utils._dask_utils import DaskIndexer
from napari.utils.colormaps import AVAILABLE_COLORMAPS
from napari.utils.events import Event
from napari.utils.events.event import WarningEmitter
from napari.utils.events.event_utils import connect_no_arg
from napari.utils.geometry import clamp_point_to_bounding_box
from napari.utils.naming import magic_name
from napari.utils.translations import trans

if TYPE_CHECKING:
    from napari.components import Dims


__all__ = ('ScalarFieldBase',)


# It is important to contain at least one abstractmethod to properly exclude this class
# in creating NAMES set inside of napari.layers.__init__
# Mixin must come before Layer
class ScalarFieldBase(Layer, ABC):
    """Base class for volumetric layers.

    Parameters
    ----------
    data : array or list of array
        Image data. Can be N >= 2 dimensional. If the last dimension has length
        3 or 4 can be interpreted as RGB or RGBA if rgb is `True`. If a
        list and arrays are decreasing in shape then the data is treated as
        a multiscale image. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    affine : n-D array or napari.utils.transforms.Affine
        (N+1, N+1) affine transformation matrix in homogeneous coordinates.
        The first (N, N) entries correspond to a linear transform and
        the final column is a length N translation vector and a 1 or a napari
        `Affine` transform object. Applied as an extra transform on top of the
        provided scale, rotate, and shear values.
    axis_labels : tuple of str, optional
        Dimension names of the layer data.
        If not provided, axis_labels will be set to (..., 'axis -2', 'axis -1').
    blending : str
        One of a list of preset blending modes that determines how RGB and
        alpha values of the layer visual get mixed. Allowed values are
        {'opaque', 'translucent', 'translucent_no_depth', 'additive', and 'minimum'}.
    cache : bool
        Whether slices of out-of-core datasets should be cached upon retrieval.
        Currently, this only applies to dask arrays.
    custom_interpolation_kernel_2d : np.ndarray
        Convolution kernel used with the 'custom' interpolation mode in 2D rendering.
    depiction : str
        3D Depiction mode. Must be one of {'volume', 'plane'}.
        The default value is 'volume'.
    experimental_clipping_planes : list of dicts, list of ClippingPlane, or ClippingPlaneList
        Each dict defines a clipping plane in 3D in data coordinates.
        Valid dictionary keys are {'position', 'normal', and 'enabled'}.
        Values on the negative side of the normal are discarded if the plane is enabled.
    metadata : dict
        Layer metadata.
    multiscale : bool
        Whether the data is a multiscale image or not. Multiscale data is
        represented by a list of array like image data. If not specified by
        the user and if the data is a list of arrays that decrease in shape
        then it will be taken to be multiscale. The first image in the list
        should be the largest. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    name : str
        Name of the layer. If not provided then will be guessed using heuristics.
    ndim : int
        Number of dimensions in the data.
    opacity : float
        Opacity of the layer visual, between 0.0 and 1.0.
    plane : dict or SlicingPlane
        Properties defining plane rendering in 3D. Properties are defined in
        data coordinates. Valid dictionary keys are
        {'position', 'normal', 'thickness', and 'enabled'}.
    projection_mode : str
        How data outside the viewed dimensions but inside the thick Dims slice will
        be projected onto the viewed dimensions. Must fit to cls._projectionclass.
    rendering : str
        Rendering mode used by vispy. Must be one of our supported
        modes.
    rotate : float, 3-tuple of float, or n-D array.
        If a float convert into a 2D rotation matrix using that value as an
        angle. If 3-tuple convert into a 3D rotation matrix, using a yaw,
        pitch, roll convention. Otherwise assume an nD rotation. Angles are
        assumed to be in degrees. They can be converted from radians with
        np.degrees if needed.
    scale : tuple of float
        Scale factors for the layer.
    shear : 1-D array or n-D array
        Either a vector of upper triangular values, or an nD shear matrix with
        ones along the main diagonal.
    translate : tuple of float
        Translation values for the layer.
    units : tuple of str or pint.Unit, optional
        Units of the layer data in world coordinates.
        If not provided, the default units are assumed to be pixels.
    visible : bool
        Whether the layer visual is currently being displayed.


    Attributes
    ----------
    data : array or list of array
        Image data. Can be N dimensional. If the last dimension has length
        3 or 4 can be interpreted as RGB or RGBA if rgb is `True`. If a list
        and arrays are decreasing in shape then the data is treated as a
        multiscale image. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    axis_labels : tuple of str
        Dimension names of the layer data.
    custom_interpolation_kernel_2d : np.ndarray
        Convolution kernel used with the 'custom' interpolation mode in 2D rendering.
    depiction : str
        3D Depiction mode used by vispy. Must be one of our supported modes.
    experimental_clipping_planes : ClippingPlaneList
        Clipping planes defined in data coordinates, used to clip the volume.
    metadata : dict
        Image metadata.
    mode : str
        Interactive mode. The normal, default mode is PAN_ZOOM, which
        allows for normal interactivity with the canvas.

        In TRANSFORM mode the image can be transformed interactively.
    multiscale : bool
        Whether the data is a multiscale image or not. Multiscale data is
        represented by a list of array like image data. The first image in the
        list should be the largest. Please note multiscale rendering is only
        supported in 2D. In 3D, only the lowest resolution scale is
        displayed.
    plane : SlicingPlane or dict
        Properties defining plane rendering in 3D. Valid dictionary keys are
        {'position', 'normal', 'thickness'}.
    rendering : str
        Rendering mode used by vispy. Must be one of our supported
        modes.
    units: tuple of pint.Unit
        Units of the layer data in world coordinates.

    Notes
    -----
    _data_view : array (N, M), (N, M, 3), or (N, M, 4)
        Image data for the currently viewed slice. Must be 2D image data, but
        can be multidimensional for RGB or RGBA images if multidimensional is
        `True`.
    """

    _colormaps = AVAILABLE_COLORMAPS
    _interpolation2d: Interpolation
    _interpolation3d: Interpolation

    def __init__(
        self,
        data,
        *,
        affine=None,
        axis_labels=None,
        blending='translucent',
        cache=True,
        custom_interpolation_kernel_2d=None,
        depiction='volume',
        experimental_clipping_planes=None,
        metadata=None,
        multiscale=None,
        name=None,
        ndim=None,
        opacity=1.0,
        plane=None,
        projection_mode='none',
        rendering='mip',
        rotate=None,
        scale=None,
        shear=None,
        translate=None,
        units=None,
        visible=True,
    ):
        if name is None and data is not None:
            name = magic_name(data)

        if isinstance(data, types.GeneratorType):
            data = list(data)

        if getattr(data, 'ndim', 2) < 2:
            raise ValueError(
                trans._('Image data must have at least 2 dimensions.')
            )

        # Determine if data is a multiscale
        self._data_raw = data
        if multiscale is None:
            multiscale, data = guess_multiscale(data)
        elif multiscale and not isinstance(data, MultiScaleData):
            data = MultiScaleData(data)

        # Determine dimensionality of the data
        if ndim is None:
            ndim = len(data.shape)

        super().__init__(
            data,
            ndim,
            affine=affine,
            axis_labels=axis_labels,
            blending=blending,
            cache=cache,
            experimental_clipping_planes=experimental_clipping_planes,
            metadata=metadata,
            multiscale=multiscale,
            name=name,
            opacity=opacity,
            projection_mode=projection_mode,
            scale=scale,
            shear=shear,
            rotate=rotate,
            translate=translate,
            units=units,
            visible=visible,
        )

        self.events.add(
            attenuation=Event,
            custom_interpolation_kernel_2d=Event,
            depiction=Event,
            interpolation=WarningEmitter(
                trans._(
                    "'layer.events.interpolation' is deprecated please use `interpolation2d` and `interpolation3d`",
                    deferred=True,
                ),
                type_name='select',
            ),
            interpolation2d=Event,
            interpolation3d=Event,
            iso_threshold=Event,
            plane=Event,
            rendering=Event,
        )

        self._array_like = True

        # Set data
        self._data = data
        if isinstance(data, MultiScaleData):
            self._data_level = len(data) - 1
            # Determine which level of the multiscale to use for the thumbnail.
            # Pick the smallest level with at least one axis >= 64. This is
            # done to prevent the thumbnail from being from one of the very
            # low resolution layers and therefore being very blurred.
            big_enough_levels = [
                np.any(np.greater_equal(p.shape, 64)) for p in data
            ]
            if np.any(big_enough_levels):
                self._thumbnail_level = np.where(big_enough_levels)[0][-1]
            else:
                self._thumbnail_level = 0
        else:
            self._data_level = 0
            self._thumbnail_level = 0
        displayed_axes = self._slice_input.displayed
        self.corner_pixels[1][displayed_axes] = (
            np.array(self.level_shapes)[self._data_level][displayed_axes] - 1
        )

        self._slice = _ImageSliceResponse.make_empty(
            slice_input=self._slice_input,
            rgb=len(self.data.shape) != self.ndim,
            dtype=self._slice_dtype(),
        )

        self._plane = SlicingPlane(thickness=1)
        # Whether to calculate clims on the next set_view_slice
        self._should_calc_clims = False
        # using self.colormap = colormap uses the setter in *derived* classes,
        # where the intention here is to use the base setter, so we use the
        # _set_colormap method. This is important for Labels layers, because
        # we don't want to use get_color before set_view_slice has been
        # triggered (self.refresh(), below).
        self.rendering = rendering
        self.depiction = depiction
        if plane is not None:
            self.plane = plane
        connect_no_arg(self.plane.events, self.events, 'plane')
        self.custom_interpolation_kernel_2d = custom_interpolation_kernel_2d

    def _post_init(self):
        # Trigger generation of view slice and thumbnail
        self.refresh()

    def _slice_dtype(self):
        """Return the dtype of the slice view.

        Overridden in Labels subclass to properly handle the
        32 and 64 bits dtypes.
        """
        return self.dtype

    @property
    def _data_view(self) -> np.ndarray:
        """Viewable image for the current slice. (compatibility)"""
        return self._slice.image.view

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def data_raw(
        self,
    ) -> LayerDataProtocol | Sequence[LayerDataProtocol]:
        """Data, exactly as provided by the user."""
        return self._data_raw

    def _get_ndim(self) -> int:
        """Determine number of dimensions of the layer."""
        return len(self.level_shapes[0])

    @property
    def _extent_data(self) -> np.ndarray:
        """Extent of layer in data coordinates.

        Returns
        -------
        extent_data : array, shape (2, D)
        """
        shape = self.level_shapes[0]
        return np.vstack([np.zeros(len(shape)), shape - 1])

    @property
    def _extent_data_augmented(self) -> np.ndarray:
        extent = self._extent_data
        return extent + [[-0.5], [+0.5]]

    @property
    def _extent_level_data(self) -> np.ndarray:
        """Extent of layer, accounting for current multiscale level, in data coordinates.

        Returns
        -------
        extent_data : array, shape (2, D)
        """
        shape = self.level_shapes[self.data_level]
        return np.vstack([np.zeros(len(shape)), shape - 1])

    @property
    def _extent_level_data_augmented(self) -> np.ndarray:
        extent = self._extent_level_data
        return extent + [[-0.5], [+0.5]]

    @property
    def data_level(self) -> int:
        """int: Current level of multiscale, or 0 if image."""
        return self._data_level

    @data_level.setter
    def data_level(self, level: int) -> None:
        if self._data_level == level:
            return
        self._data_level = level
        self.refresh(extent=False)

    def _get_level_shapes(self):
        data = self.data
        if isinstance(data, MultiScaleData):
            shapes = data.shapes
        else:
            shapes = [self.data.shape]
        return shapes

    @property
    def level_shapes(self) -> np.ndarray:
        """array: Shapes of each level of the multiscale or just of image."""
        return np.array(self._get_level_shapes())

    @property
    def downsample_factors(self) -> np.ndarray:
        """list: Downsample factors for each level of the multiscale."""
        return np.divide(self.level_shapes[0], self.level_shapes)

    @property
    def depiction(self):
        """The current 3D depiction mode.

        Selects a preset depiction mode in vispy
            * volume: images are rendered as 3D volumes.
            * plane: images are rendered as 2D planes embedded in 3D.
                plane position, normal, and thickness are attributes of
                layer.plane which can be modified directly.
        """
        return str(self._depiction)

    @depiction.setter
    def depiction(self, depiction: str | VolumeDepiction) -> None:
        """Set the current 3D depiction mode."""
        self._depiction = VolumeDepiction(depiction)
        self._update_plane_callbacks()
        self.events.depiction()

    def _reset_plane_parameters(self):
        """Set plane attributes to something valid."""
        self.plane.position = np.array(self.data.shape) / 2
        self.plane.normal = (1, 0, 0)

    def _update_plane_callbacks(self):
        """Set plane callbacks depending on depiction mode."""
        plane_drag_callback_connected = (
            plane_drag_callback in self.mouse_drag_callbacks
        )
        double_click_callback_connected = (
            plane_double_click_callback in self.mouse_double_click_callbacks
        )
        if self.depiction == VolumeDepiction.VOLUME:
            if plane_drag_callback_connected:
                self.mouse_drag_callbacks.remove(plane_drag_callback)
            if double_click_callback_connected:
                self.mouse_double_click_callbacks.remove(
                    plane_double_click_callback
                )
        elif self.depiction == VolumeDepiction.PLANE:
            if not plane_drag_callback_connected:
                self.mouse_drag_callbacks.append(plane_drag_callback)
            if not double_click_callback_connected:
                self.mouse_double_click_callbacks.append(
                    plane_double_click_callback
                )

    @property
    def plane(self):
        return self._plane

    @plane.setter
    def plane(self, value: dict | SlicingPlane) -> None:
        self._plane.update(value)
        self.events.plane()

    @property
    def custom_interpolation_kernel_2d(self):
        return self._custom_interpolation_kernel_2d

    @custom_interpolation_kernel_2d.setter
    def custom_interpolation_kernel_2d(self, value):
        if value is None:
            value = [[1]]
        self._custom_interpolation_kernel_2d = np.array(value, np.float32)
        self.events.custom_interpolation_kernel_2d()

    @abstractmethod
    def _raw_to_displayed(self, raw: np.ndarray) -> np.ndarray:
        """Determine displayed image from raw image.

        For normal image layers, just return the actual image.

        Parameters
        ----------
        raw : array
            Raw array.

        Returns
        -------
        image : array
            Displayed array.
        """
        raise NotImplementedError

    def _set_view_slice(self) -> None:
        """Set the slice output based on this layer's current state."""
        # The new slicing code makes a request from the existing state and
        # executes the request on the calling thread directly.
        # For async slicing, the calling thread will not be the main thread.
        request = self._make_slice_request_internal(
            slice_input=self._slice_input,
            data_slice=self._data_slice,
            dask_indexer=nullcontext,
        )
        response = request()
        self._update_slice_response(response)

    def _make_slice_request(self, dims: Dims) -> _ImageSliceRequest:
        """Make an image slice request based on the given dims and this image."""
        slice_input = self._make_slice_input(dims)
        # For the existing sync slicing, indices is passed through
        # to avoid some performance issues related to the evaluation of the
        # data-to-world transform and its inverse. Async slicing currently
        # absorbs these performance issues here, but we can likely improve
        # things either by caching the world-to-data transform on the layer
        # or by lazily evaluating it in the slice task itself.
        indices = slice_input.data_slice(self._data_to_world.inverse)
        return self._make_slice_request_internal(
            slice_input=slice_input,
            data_slice=indices,
            dask_indexer=self.dask_optimized_slicing,
        )

    def _make_slice_request_internal(
        self,
        *,
        slice_input: _SliceInput,
        data_slice: _ThickNDSlice,
        dask_indexer: DaskIndexer,
    ) -> _ImageSliceRequest:
        """Needed to support old-style sync slicing through _slice_dims and
        _set_view_slice.

        This is temporary scaffolding that should go away once we have completed
        the async slicing project: https://github.com/napari/napari/issues/4795
        """
        return _ImageSliceRequest(
            slice_input=slice_input,
            data=self.data,
            dask_indexer=dask_indexer,
            data_slice=data_slice,
            projection_mode=self.projection_mode,
            multiscale=self.multiscale,
            corner_pixels=self.corner_pixels,
            rgb=len(self.data.shape) != self.ndim,
            data_level=self.data_level,
            thumbnail_level=self._thumbnail_level,
            level_shapes=self.level_shapes,
            downsample_factors=self.downsample_factors,
        )

    def _update_slice_response(self, response: _ImageSliceResponse) -> None:
        """Update the slice output state currently on the layer. Currently used
        for both sync and async slicing.
        """
        response = response.to_displayed(self._raw_to_displayed)
        # We call to_displayed here to ensure that if the contrast limits
        # are outside the range of supported by vispy, then data view is
        # rescaled to fit within the range.
        self._slice_input = response.slice_input
        self._transforms[0] = response.tile_to_data
        self._slice = response

    def _get_value(self, position):
        """Value of the data at a position in data coordinates.

        Parameters
        ----------
        position : tuple
            Position in data coordinates.

        Returns
        -------
        value : tuple
            Value of the data.
        """
        if self.multiscale:
            # for multiscale data map the coordinate from the data back to
            # the tile
            coord = self._transforms['tile2data'].inverse(position)
        else:
            coord = position

        coord = np.round(coord).astype(int)

        raw = self._slice.image.raw
        shape = (
            raw.shape[:-1] if self.ndim != len(self._data.shape) else raw.shape
        )

        if self.ndim < len(coord):
            # handle 3D views of 2D data by omitting extra coordinate
            offset = len(coord) - len(shape)
            coord = coord[[d + offset for d in self._slice_input.displayed]]
        else:
            coord = coord[self._slice_input.displayed]

        if all(0 <= c < s for c, s in zip(coord, shape, strict=False)):
            value = raw[tuple(coord)]
        else:
            value = None

        if self.multiscale:
            value = (self.data_level, value)

        return value

    def _get_value_ray(
        self,
        start_point: np.ndarray | None,
        end_point: np.ndarray | None,
        dims_displayed: list[int],
    ) -> int | None:
        """Get the first non-background value encountered along a ray.

        Parameters
        ----------
        start_point : np.ndarray
            (n,) array containing the start point of the ray in data coordinates.
        end_point : np.ndarray
            (n,) array containing the end point of the ray in data coordinates.
        dims_displayed : List[int]
            The indices of the dimensions currently displayed in the viewer.

        Returns
        -------
        value : Optional[int]
            The first non-background value encountered along the ray. If none
            was encountered or the viewer is in 2D mode, returns None.
        """
        if start_point is None or end_point is None:
            return None
        if len(dims_displayed) == 3:
            # only use get_value_ray on 3D for now
            # we use dims_displayed because the image slice
            # has its dimensions  in th same order as the vispy
            # Volume
            # Account for downsampling in the case of multiscale
            # -1 means lowest resolution here.
            start_point = (
                start_point[dims_displayed]
                / self.downsample_factors[-1][dims_displayed]
            )
            end_point = (
                end_point[dims_displayed]
                / self.downsample_factors[-1][dims_displayed]
            )
            start_point = cast(np.ndarray, start_point)
            end_point = cast(np.ndarray, end_point)
            sample_ray = end_point - start_point
            length_sample_vector = np.linalg.norm(sample_ray)
            n_points = int(2 * length_sample_vector)
            sample_points = np.linspace(
                start_point, end_point, n_points, endpoint=True
            )
            im_slice = self._slice.image.raw
            # ensure the bounding box is for the proper multiscale level
            bounding_box = self._display_bounding_box_at_level(
                dims_displayed, self.data_level
            )
            # the display bounding box is returned as a closed interval
            # (i.e. the endpoint is included) by the method, but we need
            # open intervals in the code that follows, so we add 1.
            bounding_box[:, 1] += 1

            clamped = clamp_point_to_bounding_box(
                sample_points,
                bounding_box,
            ).astype(int)
            values = im_slice[tuple(clamped.T)]
            return self._calculate_value_from_ray(values)

        return None

    @abstractmethod
    def _calculate_value_from_ray(self, values):
        raise NotImplementedError

    def _get_value_3d(
        self,
        start_point: np.ndarray | None,
        end_point: np.ndarray | None,
        dims_displayed: list[int],
    ) -> int | None | tuple[int, int | None]:
        """Get the first non-background value encountered along a ray.

        Parameters
        ----------
        start_point : np.ndarray
            (n,) array containing the start point of the ray in data coordinates.
        end_point : np.ndarray
            (n,) array containing the end point of the ray in data coordinates.
        dims_displayed : List[int]
            The indices of the dimensions currently displayed in the viewer.

        Returns
        -------
        value : int or tuple
            The first non-zero value encountered along the ray. If a
            non-zero value is not encountered, returns None.
            If multiscale is True, returns a tuple of (data_level, value).
        """
        value = self._get_value_ray(
            start_point=start_point,
            end_point=end_point,
            dims_displayed=dims_displayed,
        )

        if self.multiscale and value is not None:
            return self.data_level, value

        return value

    def _get_offset_data_position(self, position: npt.NDArray) -> npt.NDArray:
        """Adjust position for offset between viewer and data coordinates.

        VisPy considers the coordinate system origin to be the canvas corner,
        while napari considers the origin to be the **center** of the corner
        pixel. To get the correct value under the mouse cursor, we need to
        shift the position by 0.5 pixels on each axis.
        """
        return position + 0.5

    def _display_bounding_box_at_level(
        self, dims_displayed: list[int], data_level: int
    ) -> npt.NDArray:
        """An axis aligned (ndisplay, 2) bounding box around the data at a given level"""
        shape = self.level_shapes[data_level]
        extent_at_level = np.vstack([np.zeros(len(shape)), shape - 1])
        return extent_at_level[:, dims_displayed].T

    def _display_bounding_box_augmented_data_level(
        self, dims_displayed: list[int]
    ) -> npt.NDArray:
        """An augmented, axis-aligned (ndisplay, 2) bounding box.
        If the layer is multiscale layer, then returns the
        bounding box of the data at the current level
        """
        return self._extent_level_data_augmented[:, dims_displayed].T



import numpy as np

from napari.benchmarks.utils import labeled_particles


def labeled_particles2d():
    seed = np.random.default_rng().integers(np.iinfo(np.int64).max)
    labels, density, points = labeled_particles(
        (1024, 1024), seed=seed, return_density=True
    )

    return [
        (density, {'name': 'density', 'metadata': {'seed': seed}}, 'image'),
        (labels, {'name': 'labels', 'metadata': {'seed': seed}}, 'labels'),
        (points, {'name': 'points', 'metadata': {'seed': seed}}, 'points'),
    ]


def labeled_particles3d():
    seed = np.random.default_rng().integers(np.iinfo(np.int64).max)
    labels, density, points = labeled_particles(
        (256, 512, 512), seed=seed, return_density=True
    )

    return [
        (density, {'name': 'density', 'metadata': {'seed': seed}}, 'image'),
        (labels, {'name': 'labels', 'metadata': {'seed': seed}}, 'labels'),
        (points, {'name': 'points', 'metadata': {'seed': seed}}, 'points'),
    ]



import math
from typing import TYPE_CHECKING

import numpy as np
from vispy.color import Colormap as VispyColormap
from vispy.gloo import Texture2D
from vispy.scene.node import Node

from napari._vispy.layers.scalar_field import (
    _DTYPE_TO_VISPY_FORMAT,
    _VISPY_FORMAT_TO_DTYPE,
    ScalarFieldLayerNode,
    VispyScalarFieldBaseLayer,
    get_dtype_from_vispy_texture_format,
)
from napari._vispy.utils.gl import get_max_texture_sizes
from napari._vispy.visuals.labels import LabelNode
from napari._vispy.visuals.volume import Volume as VolumeNode
from napari.utils.colormaps.colormap import (
    CyclicLabelColormap,
    _texture_dtype,
)

if TYPE_CHECKING:
    from napari.layers import Labels


ColorTuple = tuple[float, float, float, float]


auto_lookup_shader_uint8 = """
uniform sampler2D texture2D_values;

vec4 sample_label_color(float t) {
    if (($use_selection) && ($selection != int(t * 255))) {
        return vec4(0);
    }
    return texture2D(
        texture2D_values,
        vec2(0.0, t)
    );
}
"""

auto_lookup_shader_uint16 = """
uniform sampler2D texture2D_values;

vec4 sample_label_color(float t) {
    // uint 16
    t = t * 65535;
    if (($use_selection) && ($selection != int(t))) {
        return vec4(0);
    }
    float v = mod(t, 256);
    float v2 = (t - v) / 256;
    return texture2D(
        texture2D_values,
        vec2((v + 0.5) / 256, (v2 + 0.5) / 256)
    );
}
"""


direct_lookup_shader = """
uniform sampler2D texture2D_values;
uniform vec2 LUT_shape;

vec4 sample_label_color(float t) {
    t = t * $scale;
    return texture2D(
        texture2D_values,
        vec2(0.0, (t + 0.5) / $color_map_size)
    );
}

"""

direct_lookup_shader_many = """
uniform sampler2D texture2D_values;
uniform vec2 LUT_shape;

vec4 sample_label_color(float t) {
    t = t * $scale;
    float row = mod(t, LUT_shape.x);
    float col = int(t / LUT_shape.x);
    return texture2D(
        texture2D_values,
        vec2((col + 0.5) / LUT_shape.y, (row + 0.5) / LUT_shape.x)
    );
}
"""


class LabelVispyColormap(VispyColormap):
    def __init__(
        self,
        colormap: CyclicLabelColormap,
        view_dtype: np.dtype,
        raw_dtype: np.dtype,
    ):
        super().__init__(
            colors=['w', 'w'], controls=None, interpolation='zero'
        )
        if view_dtype.itemsize == 1:
            shader = auto_lookup_shader_uint8
        elif view_dtype.itemsize == 2:
            shader = auto_lookup_shader_uint16
        else:
            # See https://github.com/napari/napari/issues/6397
            # Using f32 dtype for textures resulted in very slow fps
            # Therefore, when we have {u,}int{8,16}, we use a texture
            # of that size, but when we have higher bits, we convert
            # to 8-bit on the CPU before sending to the shader.
            # It should thus be impossible to reach this condition.
            raise ValueError(  # pragma: no cover
                f'Cannot use dtype {view_dtype} with LabelVispyColormap'
            )

        selection = colormap._selection_as_minimum_dtype(raw_dtype)

        self.glsl_map = (
            shader.replace('$color_map_size', str(len(colormap.colors)))
            .replace('$use_selection', str(colormap.use_selection).lower())
            .replace('$selection', str(selection))
        )


class DirectLabelVispyColormap(VispyColormap):
    def __init__(
        self,
        use_selection=False,
        selection=0.0,
        scale=1.0,
        color_map_size=255,
        multi=False,
    ):
        colors = ['w', 'w']  # dummy values, since we use our own machinery
        super().__init__(colors, controls=None, interpolation='zero')
        shader = direct_lookup_shader_many if multi else direct_lookup_shader
        self.glsl_map = (
            shader.replace('$use_selection', str(use_selection).lower())
            .replace('$selection', str(selection))
            .replace('$scale', str(scale))
            .replace('$color_map_size', str(color_map_size))
        )


def build_textures_from_dict(
    color_dict: dict[int, ColorTuple], max_size: int
) -> np.ndarray:
    """This code assumes that the keys in the color_dict are sequential from 0.

    If any keys are larger than the size of the dictionary, they will
    overwrite earlier keys in the best case, or it might just crash.
    """
    if len(color_dict) > 2**23:
        raise ValueError(  # pragma: no cover
            'Cannot map more than 2**23 colors because of float32 precision. '
            f'Got {len(color_dict)}'
        )
    if len(color_dict) > max_size**2:
        raise ValueError(
            'Cannot create a 2D texture holding more than '
            f'{max_size}**2={max_size**2} colors.'
            f'Got {len(color_dict)}'
        )
    data = np.zeros(
        (
            min(len(color_dict), max_size),
            math.ceil(len(color_dict) / max_size),
            4,
        ),
        dtype=np.float32,
    )
    for key, value in color_dict.items():
        data[key % data.shape[0], key // data.shape[0]] = value
    return data


def _select_colormap_texture(
    colormap: CyclicLabelColormap, view_dtype, raw_dtype
) -> np.ndarray:
    if raw_dtype.itemsize > 2:
        color_texture = colormap._get_mapping_from_cache(view_dtype)
    else:
        color_texture = colormap._get_mapping_from_cache(raw_dtype)

    if color_texture is None:
        raise ValueError(  # pragma: no cover
            f'Cannot build a texture for dtype {raw_dtype=} and {view_dtype=}'
        )
    return color_texture.reshape(256, -1, 4)


class VispyLabelsLayer(VispyScalarFieldBaseLayer):
    layer: 'Labels'

    def __init__(self, layer, node=None, texture_format='r8') -> None:
        super().__init__(
            layer,
            node=node,
            texture_format=texture_format,
            layer_node_class=LabelLayerNode,
        )

        self.layer.events.labels_update.connect(self._on_partial_labels_update)
        self.layer.events.selected_label.connect(self._on_colormap_change)
        self.layer.events.show_selected_label.connect(self._on_colormap_change)
        self.layer.events.iso_gradient_mode.connect(
            self._on_iso_gradient_mode_change
        )
        self.layer.events.data.connect(self._on_colormap_change)
        # as we generate colormap texture based on the data type, we need to
        # update it when the data type changes

    def _on_rendering_change(self):
        # overriding the Image method, so we can maintain the same old rendering name
        if isinstance(self.node, VolumeNode):
            rendering = self.layer.rendering
            self.node.method = (
                rendering
                if rendering != 'translucent'
                else 'translucent_categorical'
            )

    def _on_colormap_change(self, event=None):
        # self.layer.colormap is a labels_colormap, which is an evented model
        # from napari.utils.colormaps.Colormap (or similar). If we use it
        # in our constructor, we have access to the texture data we need
        if (
            event is not None
            and event.type == 'selected_label'
            and not self.layer.show_selected_label
        ):
            return
        colormap = self.layer.colormap
        auto_mode = isinstance(colormap, CyclicLabelColormap)
        view_dtype = self.layer._slice.image.view.dtype
        raw_dtype = self.layer._slice.image.raw.dtype
        if auto_mode or raw_dtype.itemsize <= 2:
            if raw_dtype.itemsize > 2:
                # If the view dtype is different from the raw dtype, it is possible
                # that background pixels are not the same value as the `background_value`.
                # For example, if raw_dtype is int8 and background_value is `-1`
                # then in view dtype uint8, the background pixels will be 255
                # For data types with more than 16 bits we always cast
                # to uint8 or uint16 and background_value is always 0 in a view array.
                # The LabelColormap is EventedModel, so we need to make
                # a copy instead of temporary overwrite the background_value
                colormap = CyclicLabelColormap(**colormap.dict())
                colormap.background_value = (
                    colormap._background_as_minimum_dtype(raw_dtype)
                )
            color_texture = _select_colormap_texture(
                colormap, view_dtype, raw_dtype
            )
            self.node.cmap = LabelVispyColormap(
                colormap, view_dtype=view_dtype, raw_dtype=raw_dtype
            )
            self.node.shared_program['texture2D_values'] = Texture2D(
                color_texture,
                internalformat='rgba32f',
                interpolation='nearest',
            )
            self.texture_data = color_texture

        elif not auto_mode:  # only for raw_dtype.itemsize > 2
            color_dict = colormap._values_mapping_to_minimum_values_set()[1]
            max_size = get_max_texture_sizes()[0]
            val_texture = build_textures_from_dict(color_dict, max_size)

            dtype = _texture_dtype(
                self.layer._direct_colormap._num_unique_colors + 2,
                raw_dtype,
            )
            if issubclass(dtype.type, np.integer):
                scale = np.iinfo(dtype).max
            else:  # float32 texture
                scale = 1.0

            self.node.cmap = DirectLabelVispyColormap(
                use_selection=colormap.use_selection,
                selection=colormap.selection,
                scale=scale,
                color_map_size=val_texture.shape[0],
                multi=val_texture.shape[1] > 1,
            )
            self.node.shared_program['texture2D_values'] = Texture2D(
                val_texture,
                internalformat='rgba32f',
                interpolation='nearest',
            )
            self.node.shared_program['LUT_shape'] = val_texture.shape[:2]
        else:
            self.node.cmap = VispyColormap(*colormap)

    def _on_iso_gradient_mode_change(self):
        if isinstance(self.node, VolumeNode):
            self.node.iso_gradient_mode = self.layer.iso_gradient_mode

    def _on_partial_labels_update(self, event):
        if not self.layer.loaded:
            return

        raw_displayed = self.layer._slice.image.raw
        ndims = len(event.offset)

        if self.node._texture.shape[:ndims] != raw_displayed.shape[:ndims]:
            # TODO: I'm confused by this whole process; should this refresh be changed?
            self.layer.refresh()
            return

        self.node._texture.scale_and_set_data(
            event.data, copy=False, offset=event.offset
        )
        self.node.update()

    def reset(self, event=None) -> None:
        super().reset()
        self._on_colormap_change()
        self._on_iso_gradient_mode_change()


class LabelLayerNode(ScalarFieldLayerNode):
    def __init__(self, custom_node: Node = None, texture_format=None):
        self._custom_node = custom_node
        self._setup_nodes(texture_format)

    def _setup_nodes(self, texture_format):
        self._image_node = LabelNode(
            (
                None
                if (texture_format is None or texture_format == 'auto')
                else np.zeros(
                    (1, 1),
                    dtype=get_dtype_from_vispy_texture_format(texture_format),
                )
            ),
            method='auto',
            texture_format=texture_format,
        )

        self._volume_node = VolumeNode(
            np.zeros(
                (1, 1, 1),
                dtype=get_dtype_from_vispy_texture_format(texture_format),
            ),
            clim=[0, 2**23 - 1],
            texture_format=texture_format,
            interpolation='nearest',
        )

    def get_node(self, ndisplay: int, dtype=None) -> Node:
        res = self._image_node if ndisplay == 2 else self._volume_node

        if (
            res.texture_format != 'auto'
            and dtype is not None
            and _VISPY_FORMAT_TO_DTYPE[res.texture_format] != dtype
        ):
            self._setup_nodes(_DTYPE_TO_VISPY_FORMAT[dtype])
            return self.get_node(ndisplay, dtype)
        return res



import contextlib
import sys
import time
from enum import Enum
from typing import TYPE_CHECKING
from unittest.mock import Mock

import numpy as np
import numpy.testing as npt
import pytest
from magicgui import magicgui

from napari import Viewer, layers, types
from napari._tests.utils import layer_test_data
from napari.layers import Image, Labels, Layer, Points, Surface
from napari.utils._magicgui import _get_ndim_from_data
from napari.utils._proxies import PublicOnlyProxy
from napari.utils.migrations import _DeprecatingDict
from napari.utils.misc import all_subclasses

if TYPE_CHECKING:
    import napari.types

try:
    import qtpy  # noqa: F401 need to be ignored as qtpy may be available but Qt bindings may not be
except ModuleNotFoundError:
    pytest.skip('Cannot test magicgui without qtpy.', allow_module_level=True)
except RuntimeError:
    pytest.skip(
        'Cannot test magicgui without Qt bindings.', allow_module_level=True
    )


@pytest.fixture
def image_layer():
    return Image(
        np.empty((40, 40), dtype=np.uint8),
        scale=(1, 2),
        translate=(3, 4),
        units=('nm', 'um'),
    )


@pytest.fixture
def image_layer3d():
    return Image(
        np.empty((10, 40, 40), dtype=np.uint8),
        scale=(1, 2, 2),
        translate=(3, 4, 4),
        units=('nm', 'um', 'um'),
    )


@pytest.fixture
def image_layer_rgb():
    return Image(
        np.empty((40, 40, 3), dtype=np.uint8),
        scale=(1, 2),
        translate=(3, 4),
        units=('nm', 'um'),
    )


@pytest.fixture
def labels_layer():
    return Labels(
        np.empty((40, 40), dtype=np.uint8),
        scale=(1, 2),
        translate=(3, 4),
        units=('nm', 'um'),
    )


@pytest.fixture
def points_layer():
    return Points(
        np.empty((10, 2), dtype=np.uint8).astype(np.float32),
        scale=(1, 2),
        translate=(3, 4),
        units=('nm', 'um'),
    )


@pytest.fixture
def surface_layer():
    rng = np.random.default_rng(0)
    return Surface(
        (20 * rng.random((10, 3)), rng.integers(10, size=(6, 3))),
        scale=(1, 2, 2),
        translate=(3, 4, 4),
        units=('nm', 'um', 'um'),
    )


@pytest.fixture(
    params=['image_layer', 'labels_layer', 'points_layer', 'image_layer_rgb']
)
def layer_and_type(request):
    data = request.getfixturevalue(request.param)
    return data, getattr(types, f'{data.__class__.__name__}Data')


# only test the first of each layer type
test_data = []
for cls in all_subclasses(Layer):
    # OctTree Image doesn't have layer_test_data
    with contextlib.suppress(StopIteration):
        test_data.append(next(x for x in layer_test_data if x[0] is cls))
test_data.sort(key=lambda x: x[0].__name__)  # required for xdist to work


@pytest.mark.parametrize(('LayerType', 'data', 'ndim'), layer_test_data)
def test_get_ndim_from_data(LayerType, data, ndim):
    assert _get_ndim_from_data(data, LayerType.__name__.lower()) == ndim


@pytest.mark.parametrize(('LayerType', 'data', 'ndim'), test_data)
def test_magicgui_add_data(make_napari_viewer, LayerType, data, ndim):
    """Test that annotating with napari.types.<layer_type>Data works.

    It expects a raw data format (like a numpy array) and will add a layer
    of the corresponding type to the viewer.
    """
    viewer = make_napari_viewer()
    dtype = getattr(types, f'{LayerType.__name__}Data')

    @magicgui
    # where `dtype` is something like napari.types.ImageData
    def add_data() -> dtype:  # type: ignore
        # and data is just the bare numpy-array or similar
        return data

    viewer.window.add_dock_widget(add_data)
    add_data()
    assert len(viewer.layers) == 1
    assert isinstance(viewer.layers[0], LayerType)
    assert viewer.layers[0].source.widget == add_data


def test_magicgui_add_data_inheritance(
    make_napari_viewer, layer_and_type, image_layer
):
    """This test validates if the scale and translate are inherited from the
    previous layer when adding a new layer with magicgui if function requests,
    a LayerData type.
    """
    layer, type_ = layer_and_type
    viewer = make_napari_viewer()
    viewer.add_layer(image_layer)

    @magicgui
    def add_data(data: types.ImageData) -> type_:
        return layer.data

    viewer.window.add_dock_widget(add_data)
    add_data()
    assert len(viewer.layers) == 2
    assert isinstance(viewer.layers[1], layer.__class__)
    npt.assert_array_equal(viewer.layers[1].scale, (1, 2))
    npt.assert_array_equal(viewer.layers[1].translate, (3, 4))
    assert viewer.layers[1].units == viewer.layers[0].units


def test_magicgui_add_data_inheritance_surface(
    make_napari_viewer, surface_layer, image_layer3d
):
    """This test validates if the scale and translate are inherited from the
    previous layer when adding a new layer with magicgui if function requests,
    a LayerData type.
    """
    viewer = make_napari_viewer()
    viewer.add_layer(image_layer3d)

    @magicgui
    def add_data(data: types.ImageData) -> types.SurfaceData:
        return surface_layer.data

    viewer.window.add_dock_widget(add_data)
    add_data()
    assert len(viewer.layers) == 2
    assert isinstance(viewer.layers[1], Surface)
    npt.assert_array_equal(viewer.layers[1].scale, (1, 2, 2))
    npt.assert_array_equal(viewer.layers[1].translate, (3, 4, 4))
    assert viewer.layers[1].units == viewer.layers[0].units


def test_magicgui_add_data_inheritance_two_layer(make_napari_viewer):
    """This test validates if the scale and translate are inherited if more than
    one source layer is passed to function when adding a new layer
    with magicgui if function requests, a LayerData type.
    """
    rng = np.random.default_rng(0)
    viewer = make_napari_viewer()
    viewer.add_image(rng.random((10, 10)), scale=(1, 2), translate=(3, 4))
    viewer.add_labels(
        (rng.random((10, 10)) > 0.5).astype('uint8'),
        scale=(1, 2),
        translate=(3, 4),
    )

    @magicgui
    def add_data(
        data1: types.ImageData, data2: types.LabelsData
    ) -> types.LabelsData:
        return (data1 * data2).astype('uint8')

    viewer.window.add_dock_widget(add_data)
    add_data()
    assert len(viewer.layers) == 3
    assert isinstance(viewer.layers[2], Labels)
    npt.assert_array_equal(viewer.layers[2].scale, (1, 2))
    npt.assert_array_equal(viewer.layers[2].translate, (3, 4))


def test_magicgui_add_data_inheritance_two_layer_inconsistent(
    make_napari_viewer, monkeypatch
):
    """This test validates the scale and translate are not inherited from the
    previous layers with inconsistend metadata when adding a new layer
    with magicgui if function requests,
    a LayerData type.
    """
    rng = np.random.default_rng(0)
    viewer = make_napari_viewer()
    viewer.add_image(rng.random((10, 10)), scale=(1, 2), translate=(3, 4))
    viewer.add_labels(
        (rng.random((10, 10)) > 0.5).astype('uint8'),
        scale=(2, 2),
        translate=(3, 4),
    )

    @magicgui
    def add_data(
        data1: types.ImageData, data2: types.LabelsData
    ) -> types.LabelsData:
        return (data1 * data2).astype('uint8')

    viewer.window.add_dock_widget(add_data)
    mock = Mock()
    monkeypatch.setattr(
        'napari.utils.notifications.notification_manager.dispatch', mock
    )
    add_data()
    mock.assert_called_once()
    assert mock.call_args[0][0].message.startswith('Cannot inherit spatial')
    assert len(viewer.layers) == 3
    assert isinstance(viewer.layers[2], Labels)
    npt.assert_array_equal(viewer.layers[2].scale, (1, 1))
    npt.assert_array_equal(viewer.layers[2].translate, (0, 0))


def test_magicgui_add_layer_inheritance(make_napari_viewer):
    """This test validates if the scale and translate are inherited from the
    previous layer when adding a new layer with magicgui if function requests,
    a Layer type.
    It also checks if the presence of additional combo box in the
    function does not affect getting the data from the previous layer.
    """
    rng = np.random.default_rng(0)
    viewer = make_napari_viewer()
    viewer.add_image(rng.random((10, 10)), scale=(2, 2), translate=(1, 1))

    class SampleEnum(Enum):
        A = 'A'
        B = 'B'

    @magicgui
    def add_data(
        data: Image, factor: float = 0.5, combo: SampleEnum = SampleEnum.A
    ) -> types.LabelsData:
        return (data.data > factor).astype(
            'uint8' if combo == SampleEnum.A else 'uint16'
        )

    viewer.window.add_dock_widget(add_data)
    add_data()
    assert len(viewer.layers) == 2
    assert isinstance(viewer.layers[1], Labels)
    npt.assert_array_equal(viewer.layers[1].scale, (2, 2))
    npt.assert_array_equal(viewer.layers[1].translate, (1, 1))


def test_magicgui_add_data_inheritance_upper_dim(make_napari_viewer):
    """In the current implementation, layer with dimensionality lower than produced data are ignored"""
    rng = np.random.default_rng(0)
    viewer = make_napari_viewer()
    viewer.add_image(rng.random((10, 10)), scale=(2, 2), translate=(1, 1))

    @magicgui
    def add_data(data: types.ImageData) -> types.LabelsData:
        return np.stack([(data > 0.5), (data > 0.4)]).astype('uint8')

    viewer.window.add_dock_widget(add_data)
    add_data()
    assert len(viewer.layers) == 2
    assert isinstance(viewer.layers[1], Labels)
    npt.assert_array_equal(viewer.layers[1].scale, (1, 1, 1))
    npt.assert_array_equal(viewer.layers[1].translate, (0, 0, 0))


def test_magicgui_add_data_inheritance_less_dim(make_napari_viewer):
    viewer = make_napari_viewer()
    viewer.add_image(
        np.random.rand(4, 10, 10), scale=(1, 2, 2), translate=(2, 1, 1)
    )

    @magicgui
    def add_data(data: types.ImageData) -> types.LabelsData:
        return (data[0] > 0.5).astype('uint8')

    viewer.window.add_dock_widget(add_data)
    add_data()
    assert len(viewer.layers) == 2
    assert isinstance(viewer.layers[1], Labels)
    npt.assert_array_equal(viewer.layers[1].scale, (2, 2))
    npt.assert_array_equal(viewer.layers[1].translate, (1, 1))


def test_add_layer_data_to_viewer_optional(make_napari_viewer):
    viewer = make_napari_viewer()

    @magicgui
    def func_optional(a: bool) -> 'napari.types.ImageData | None':
        if a:
            return np.zeros((10, 10))
        return None

    viewer.window.add_dock_widget(func_optional)
    assert not viewer.layers

    func_optional(a=True)

    assert len(viewer.layers) == 1

    func_optional(a=False)

    assert len(viewer.layers) == 1


@pytest.mark.parametrize(('LayerType', 'data', 'ndim'), test_data)
def test_magicgui_add_future_data(
    qtbot, make_napari_viewer, LayerType, data, ndim
):
    """Test that annotating with Future[] works."""
    from concurrent.futures import Future
    from functools import partial

    from qtpy.QtCore import QTimer

    viewer = make_napari_viewer()
    dtype = getattr(types, f'{LayerType.__name__}Data')

    @magicgui
    # where `dtype` is something like napari.types.ImageData
    def add_data() -> Future[dtype]:  # type: ignore
        future = Future()
        # simulate something that isn't immediately ready when function returns
        QTimer.singleShot(10, partial(future.set_result, data))
        return future

    viewer.window.add_dock_widget(add_data)

    def _assert_stuff():
        assert len(viewer.layers) == 1
        assert isinstance(viewer.layers[0], LayerType)
        assert viewer.layers[0].source.widget == add_data

    assert len(viewer.layers) == 0
    with qtbot.waitSignal(viewer.layers.events.inserted):
        add_data()
    _assert_stuff()


def test_magicgui_add_threadworker(qtbot, make_napari_viewer):
    """Test that annotating with FunctionWorker works."""
    from napari.qt.threading import FunctionWorker, thread_worker

    viewer = make_napari_viewer()
    DATA = np.random.rand(10, 10)

    @magicgui
    def add_data(x: int) -> FunctionWorker[types.ImageData]:
        @thread_worker(start_thread=False)
        def _slow():
            time.sleep(0.1)
            return DATA

        return _slow()

    viewer.window.add_dock_widget(add_data)

    assert len(viewer.layers) == 0
    worker = add_data()
    # normally you wouldn't start the worker outside of the mgui function
    # this is just to make testing with threads easier
    with qtbot.waitSignal(worker.finished):
        worker.start()

    assert len(viewer.layers) == 1
    assert isinstance(viewer.layers[0], Image)
    assert viewer.layers[0].source.widget == add_data
    assert np.array_equal(viewer.layers[0].data, DATA)


@pytest.mark.parametrize(('LayerType', 'data', 'ndim'), test_data)
def test_magicgui_get_data(make_napari_viewer, LayerType, data, ndim):
    """Test that annotating parameters with napari.types.<layer_type>Data.

    This will provide the same dropdown menu appearance as when annotating
    a parameter with napari.layers.<layer_type>... but the function will
    receive `layer.data` rather than `layer`
    """
    viewer = make_napari_viewer()
    dtype = getattr(types, f'{LayerType.__name__}Data')

    @magicgui
    # where `dtype` is something like napari.types.ImageData
    def add_data(x: dtype):
        # and data is just the bare numpy-array or similar
        return data

    viewer.window.add_dock_widget(add_data)
    layer = LayerType(data)
    viewer.add_layer(layer)


@pytest.mark.parametrize(('LayerType', 'data', 'ndim'), test_data)
def test_magicgui_add_layer(make_napari_viewer, LayerType, data, ndim):
    viewer = make_napari_viewer()

    @magicgui
    def add_layer() -> LayerType:
        return LayerType(data)

    viewer.window.add_dock_widget(add_layer)
    add_layer()
    assert len(viewer.layers) == 1
    assert isinstance(viewer.layers[0], LayerType)
    assert viewer.layers[0].source.widget == add_layer


def test_magicgui_add_layer_list(make_napari_viewer):
    viewer = make_napari_viewer()

    @magicgui
    def add_layer() -> list[Layer]:
        a = Image(data=np.random.randint(0, 10, size=(10, 10)))
        b = Labels(data=np.random.randint(0, 10, size=(10, 10)))
        return [a, b]

    viewer.window.add_dock_widget(add_layer)
    add_layer()
    assert len(viewer.layers) == 2
    assert isinstance(viewer.layers[0], Image)
    assert isinstance(viewer.layers[1], Labels)

    assert viewer.layers[0].source.widget == add_layer
    assert viewer.layers[1].source.widget == add_layer


def test_magicgui_add_layer_data_tuple(make_napari_viewer):
    viewer = make_napari_viewer()

    @magicgui
    def add_layer() -> types.LayerDataTuple:
        data = (
            np.random.randint(0, 10, size=(10, 10)),
            {'name': 'hi'},
            'labels',
        )
        # it works fine to just return `data`
        # but this will avoid mypy/linter errors and has no runtime burden
        return types.LayerDataTuple(data)

    viewer.window.add_dock_widget(add_layer)
    add_layer()
    assert len(viewer.layers) == 1
    assert isinstance(viewer.layers[0], Labels)
    assert viewer.layers[0].source.widget == add_layer


def test_magicgui_add_layer_data_tuple_list(make_napari_viewer):
    viewer = make_napari_viewer()

    @magicgui
    def add_layer() -> list[types.LayerDataTuple]:
        data1 = (np.random.rand(10, 10), {'name': 'hi'})
        data2 = (
            np.random.randint(0, 10, size=(10, 10)),
            {'name': 'hi2'},
            'labels',
        )
        return [data1, data2]  # type: ignore

    viewer.window.add_dock_widget(add_layer)
    add_layer()
    assert len(viewer.layers) == 2
    assert isinstance(viewer.layers[0], Image)
    assert isinstance(viewer.layers[1], Labels)

    assert viewer.layers[0].source.widget == add_layer
    assert viewer.layers[1].source.widget == add_layer


def test_magicgui_data_updated(make_napari_viewer):
    """Test that magic data parameters stay up to date."""
    viewer = make_napari_viewer()

    _returns = []  # the value of x returned from func

    @magicgui(auto_call=True)
    def func(x: types.PointsData):
        _returns.append(x)

    viewer.window.add_dock_widget(func)
    points = viewer.add_points(None)
    # func will have been called with an empty points
    np.testing.assert_allclose(_returns[-1], np.empty((0, 2)))
    points.add((10, 10))
    # func will have been called with 1 data including 1 point
    np.testing.assert_allclose(_returns[-1], np.array([[10, 10]]))
    points.add((15, 15))
    # func will have been called with 1 data including 2 points
    np.testing.assert_allclose(_returns[-1], np.array([[10, 10], [15, 15]]))


def test_magicgui_get_viewer(make_napari_viewer):
    """Test that annotating with napari.Viewer gets the Viewer"""
    # Make two DIFFERENT viewers
    viewer1 = make_napari_viewer()
    viewer2 = make_napari_viewer()
    assert viewer2 is not viewer1
    # Ensure one is returned by napari.current_viewer()
    from napari import current_viewer

    assert current_viewer() is viewer2

    @magicgui
    def func(v: Viewer):
        return v

    def func_returns(v: Viewer) -> bool:
        """Helper function determining whether func() returns v"""
        func_viewer = func()
        assert isinstance(func_viewer, PublicOnlyProxy)
        return func_viewer.__wrapped__ is v

    # We expect func's Viewer to be current_viewer, not viewer
    assert func_returns(viewer2)
    assert not func_returns(viewer1)
    # With viewer as parent, it should be returned instead
    viewer1.window.add_dock_widget(func)
    assert func_returns(viewer1)
    assert not func_returns(viewer2)
    # no widget should be shown
    assert not func.v.visible
    # ensure that viewer2 is still the current viewer
    assert current_viewer() is viewer2


MGUI_EXPORTS = ['napari.layers.Layer', 'napari.Viewer']
MGUI_EXPORTS += [f'napari.types.{nm.title()}Data' for nm in layers.NAMES]
NAMES = ('Image', 'Labels', 'Layer', 'Points', 'Shapes', 'Surface')


@pytest.mark.parametrize('name', sorted(MGUI_EXPORTS))
def test_mgui_forward_refs(name, monkeypatch):
    """make sure that magicgui's `get_widget_class` returns the right widget type
    for the various napari types... even when expressed as strings.
    """
    import magicgui.widgets
    from magicgui.type_map import get_widget_class

    monkeypatch.delitem(sys.modules, 'napari')
    monkeypatch.delitem(sys.modules, 'napari.viewer')
    monkeypatch.delitem(sys.modules, 'napari.types')
    monkeypatch.setattr(
        'napari.utils.action_manager.action_manager._actions', {}
    )
    # need to clear all of these submodules too, otherwise the layers are oddly not
    # subclasses of napari.layers.Layer, and napari.layers.NAMES
    # oddly ends up as an empty set
    for m in list(sys.modules):
        if m.startswith('napari.layers') and 'utils' not in m:
            monkeypatch.delitem(sys.modules, m)

    wdg, options = get_widget_class(annotation=name)
    if name == 'napari.Viewer':
        assert wdg == magicgui.widgets.EmptyWidget
        assert 'bind' in options
    else:
        assert wdg == magicgui.widgets.Combobox


def test_layers_populate_immediately(make_napari_viewer):
    """make sure that the layers dropdown is populated upon adding to viewer"""
    from magicgui.widgets import create_widget

    labels_layer = create_widget(annotation=Labels, label='ROI')
    viewer = make_napari_viewer()
    viewer.add_labels(np.zeros((10, 10), dtype=int))
    assert not len(labels_layer.choices)
    viewer.window.add_dock_widget(labels_layer)
    assert len(labels_layer.choices) == 1


def test_from_layer_data_tuple_accept_deprecating_dict(make_napari_viewer):
    """Test that a function returning a layer data tuple runs without error."""
    viewer = make_napari_viewer()

    @magicgui
    def from_layer_data_tuple() -> types.LayerDataTuple:
        data = np.zeros((10, 10))
        meta = _DeprecatingDict({'name': 'test_image'})
        layer_type = 'image'
        return data, meta, layer_type

    viewer.window.add_dock_widget(from_layer_data_tuple)
    from_layer_data_tuple()
    assert len(viewer.layers) == 1
    assert isinstance(viewer.layers[0], Image)
    assert viewer.layers[0].name == 'test_image'



import numpy as np
import pytest

from napari._vispy.layers.vectors import (
    generate_vector_meshes,
    generate_vector_meshes_2D,
)


@pytest.mark.parametrize(
    ('edge_width', 'length', 'dims', 'style'),
    [
        (0, 0, 2, 'line'),
        (0.3, 0.3, 2, 'line'),
        (1, 1, 3, 'line'),
        (0, 0, 2, 'triangle'),
        (0.3, 0.3, 2, 'triangle'),
        (1, 1, 3, 'triangle'),
        (0, 0, 2, 'arrow'),
        (0.3, 0.3, 2, 'arrow'),
        (1, 1, 3, 'arrow'),
    ],
)
def test_generate_vector_meshes(edge_width, length, dims, style):
    n = 10

    data = np.random.random((n, 2, dims))
    vertices, faces = generate_vector_meshes(
        data, width=edge_width, length=length, vector_style=style
    )
    vertices_length, vertices_dims = vertices.shape
    faces_length, faces_dims = faces.shape

    if dims == 2:
        if style == 'line':
            assert vertices_length == 4 * n
            assert faces_length == 2 * n
        elif style == 'triangle':
            assert vertices_length == 3 * n
            assert faces_length == n
        elif style == 'arrow':
            assert vertices_length == 7 * n
            assert faces_length == 3 * n

    elif dims == 3:
        if style == 'line':
            assert vertices_length == 8 * n
            assert faces_length == 4 * n
        elif style == 'triangle':
            assert vertices_length == 6 * n
            assert faces_length == 2 * n
        elif style == 'arrow':
            assert vertices_length == 14 * n
            assert faces_length == 6 * n

    assert vertices_dims == dims
    assert faces_dims == 3


@pytest.mark.parametrize(
    ('edge_width', 'length', 'style', 'p'),
    [
        (0, 0, 'line', (1, 0, 0)),
        (0.3, 0.3, 'line', (0, 1, 0)),
        (1, 1, 'line', (0, 0, 1)),
        (0, 0, 'triangle', (1, 0, 0)),
        (0.3, 0.3, 'triangle', (0, 1, 0)),
        (1, 1, 'triangle', (0, 0, 1)),
        (0, 0, 'arrow', (1, 0, 0)),
        (0.3, 0.3, 'arrow', (0, 1, 0)),
        (1, 1, 'arrow', (0, 0, 1)),
    ],
)
def test_generate_vector_meshes_2D(edge_width, length, style, p):
    n = 10
    dims = 2

    data = np.random.random((n, 2, dims))
    vertices, faces = generate_vector_meshes_2D(
        data, width=edge_width, length=length, vector_style=style, p=p
    )
    vertices_length, vertices_dims = vertices.shape
    faces_length, faces_dims = faces.shape

    if style == 'line':
        assert vertices_length == 4 * n
        assert faces_length == 2 * n
    elif style == 'triangle':
        assert vertices_length == 3 * n
        assert faces_length == n
    elif style == 'arrow':
        assert vertices_length == 7 * n
        assert faces_length == 3 * n

    assert vertices_dims == dims
    assert faces_dims == 3


@pytest.mark.parametrize(
    ('initial_vector_style', 'new_vector_style'),
    [
        ('line', 'line'),
        ('line', 'triangle'),
        ('line', 'arrow'),
        ('triangle', 'line'),
        ('triangle', 'triangle'),
        ('triangle', 'arrow'),
        ('arrow', 'line'),
        ('arrow', 'triangle'),
        ('arrow', 'arrow'),
    ],
)
def test_vector_style_change(
    make_napari_viewer, initial_vector_style, new_vector_style
):
    # initialize viewer
    viewer = make_napari_viewer()
    # add a vector layer
    vector_layer = viewer.add_vectors(
        vector_style=initial_vector_style, name='vectors'
    )

    class Counter:
        def __init__(self):
            self.count = 0

        def increment_count(self, event):
            self.count += 1

    # initialize counter
    counter = Counter()
    # connect counter to vector_style change
    vector_layer.events.vector_style.connect(counter.increment_count)

    # change vector_style
    vector_layer.vector_style = new_vector_style

    # check if counter was called
    if initial_vector_style == new_vector_style:
        assert counter.count == 0
    else:
        assert counter.count == 1



import logging

from qtpy.QtCore import Qt
from qtpy.QtGui import QFontDatabase
from qtpy.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from napari._qt.dialogs.qt_about import QtCopyToClipboardButton
from napari.utils._logging import LOG_STREAM, get_log_level_value
from napari.utils.translations import trans


class LogWidget(QWidget):
    """
    Widget for inspecting and filtering logging output.
    """

    def __init__(
        self,
        parent=None,
    ) -> None:
        super().__init__(parent)

        self.layout = QVBoxLayout()

        # Description
        title_label = QLabel(trans._('logger'))
        title_label.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse
        )
        self.layout.addWidget(title_label)

        # level selection
        self.level_layout = QHBoxLayout()
        self.level_layout.addWidget(QLabel('Current log level:'))
        self.loglevel = QComboBox()
        self.loglevel.addItems(
            ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
        )
        self.level_layout.addWidget(self.loglevel)
        self.level_layout.addStretch()
        self.layout.addLayout(self.level_layout)

        # filtering
        self.filter_layout = QHBoxLayout()
        self.filter_layout.addWidget(QLabel('Filter:'))
        self.level_filter = QComboBox()
        self.level_filter.addItems(
            ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
        )
        self.filter_layout.addWidget(self.level_filter)

        self.text_filter = QLineEdit()
        self.text_filter.setPlaceholderText('text filter')
        self.filter_layout.addWidget(self.text_filter)
        self.layout.addLayout(self.filter_layout)

        # log text box
        self.log_text_box = QTextEdit()
        self.log_text_box.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse
        )
        self.log_text_box.setLineWrapMode(QTextEdit.NoWrap)
        font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        self.log_text_box.setFont(font)

        # Add text copy button
        self.infoCopyButton = QtCopyToClipboardButton(self.log_text_box)
        self.info_layout = QHBoxLayout()
        self.info_layout.addWidget(self.log_text_box, 1)
        self.info_layout.addWidget(
            self.infoCopyButton, 0, Qt.AlignmentFlag.AlignTop
        )
        self.info_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.layout.addLayout(self.info_layout)

        self.setLayout(self.layout)

        self.loglevel.currentTextChanged.connect(self._on_loglevel_change)
        self.level_filter.currentTextChanged.connect(self._on_change)
        self.text_filter.textChanged.connect(self._on_change)

        # TODO: super laggy when open :/
        LOG_STREAM.changed.connect(self._on_new_message)
        self._prev_pos = None
        self.log_text_box.verticalScrollBar().rangeChanged.connect(
            self._jump_to_pos
        )

        self._on_loglevel_change()
        self._on_change()
        self._jump_to_pos()

    def _on_loglevel_change(self, event=None):
        level = get_log_level_value(event)
        logging.getLogger().setLevel(level)

    def _on_new_message(self, event=None):
        self._prev_pos = self._scroll_pos()

        log = LOG_STREAM.get_filtered_logs_html(
            self.level_filter.currentText(),
            self.text_filter.text(),
            last_only=True,
        )[0]
        self.log_text_box.append(log)

    def _on_change(self, event=None):
        self._prev_pos = self._scroll_pos()

        logs = LOG_STREAM.get_filtered_logs_html(
            self.level_filter.currentText(), self.text_filter.text()
        )
        self.log_text_box.clear()
        for log in logs:
            # by looping here instead of joinging the lines,
            # we ensure each line is separate (allows better selection)
            self.log_text_box.append(log)

    def _jump_to_pos(self, event=None):
        # maintains position when updating the contents of the text
        # for some reason using scrollbar.setValue() doesn't keep up,
        # cursor's better
        scrollbar = self.log_text_box.verticalScrollBar()
        if self._prev_pos is None:
            scrollbar.setValue(scrollbar.maximum())
        else:
            scrollbar.setValue(self._prev_pos)

        # self.log_text_box.moveCursor(self.log_text_box.textCursor().End)
        # self.log_text_box.moveCursor(self.log_text_box.textCursor().StartOfLine)
        # self.log_text_box.ensureCursorVisible()

    def _scroll_pos(self):
        scrollbar = self.log_text_box.verticalScrollBar()
        curr = scrollbar.value()
        if curr == scrollbar.maximum():
            return None
        return curr



import contextlib
from typing import TYPE_CHECKING

import numpy as np
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
)
from superqt import QLabeledSlider

from napari._qt.layer_controls.qt_layer_controls_base import QtLayerControls
from napari._qt.utils import qt_signals_blocked
from napari._qt.widgets.qt_color_swatch import QColorSwatchEdit
from napari._qt.widgets.qt_mode_buttons import QtModePushButton
from napari.layers.points._points_constants import (
    SYMBOL_TRANSLATION,
    SYMBOL_TRANSLATION_INVERTED,
    Mode,
)
from napari.utils.action_manager import action_manager
from napari.utils.events import disconnect_events
from napari.utils.translations import trans

if TYPE_CHECKING:
    import napari.layers


class QtPointsControls(QtLayerControls):
    """Qt view and controls for the napari Points layer.

    Parameters
    ----------
    layer : napari.layers.Points
        An instance of a napari Points layer.

    Attributes
    ----------
    layer : napari.layers.Points
        An instance of a napari Points layer.
    MODE : Enum
        Available modes in the associated layer.
    PAN_ZOOM_ACTION_NAME : str
        String id for the pan-zoom action to bind to the pan_zoom button.
    TRANSFORM_ACTION_NAME : str
        String id for the transform action to bind to the transform button.
    addition_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to add points to layer.
    button_group : qtpy.QtWidgets.QButtonGroup
        Button group of points layer modes (ADD, PAN_ZOOM, SELECT).
    delete_button : qtpy.QtWidgets.QtModePushButton
        Button to delete points from layer.
    borderColorEdit : QColorSwatchEdit
        Widget to select display color for points borders.
    faceColorEdit : QColorSwatchEdit
        Widget to select display color for points faces.
    outOfSliceCheckBox : qtpy.QtWidgets.QCheckBox
        Checkbox to indicate whether to render out of slice.
    panzoom_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to activate move camera mode for layer.
    transform_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select transform mode.
    select_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select points from layer.
    sizeSlider : qtpy.QtWidgets.QSlider
        Slider controlling size of points.
    symbolComboBox : qtpy.QtWidgets.QComboBox
        Drop down list of symbol options for points markers.

    Raises
    ------
    ValueError
        Raise error if points mode is not recognized.
        Points mode must be one of: ADD, PAN_ZOOM, or SELECT.
    """

    layer: 'napari.layers.Points'
    MODE = Mode
    PAN_ZOOM_ACTION_NAME = 'activate_points_pan_zoom_mode'
    TRANSFORM_ACTION_NAME = 'activate_points_transform_mode'

    def __init__(self, layer) -> None:
        super().__init__(layer)

        self.layer.events.out_of_slice_display.connect(
            self._on_out_of_slice_display_change
        )
        self.layer.events.symbol.connect(self._on_current_symbol_change)
        self.layer.events.size.connect(self._on_current_size_change)
        self.layer.events.current_size.connect(self._on_current_size_change)
        self.layer.events.current_border_color.connect(
            self._on_current_border_color_change
        )
        self.layer._border.events.current_color.connect(
            self._on_current_border_color_change
        )
        self.layer.events.current_face_color.connect(
            self._on_current_face_color_change
        )
        self.layer._face.events.current_color.connect(
            self._on_current_face_color_change
        )
        self.layer.events.current_symbol.connect(
            self._on_current_symbol_change
        )
        self.layer.text.events.visible.connect(self._on_text_visibility_change)

        sld = QLabeledSlider(Qt.Orientation.Horizontal)
        sld.setToolTip(
            trans._(
                'Change the size of currently selected points and any added afterwards.'
            )
        )
        sld.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        sld.setMinimum(1)
        if self.layer.size.size:
            max_value = max(100, int(np.max(self.layer.size)) + 1)
        else:
            max_value = 100
        sld.setMaximum(max_value)
        sld.setSingleStep(1)
        value = self.layer.current_size
        sld.setValue(int(value))
        sld.valueChanged.connect(self.changeCurrentSize)
        self.sizeSlider = sld

        self.faceColorEdit = QColorSwatchEdit(
            initial_color=self.layer.current_face_color,
            tooltip=trans._(
                'Click to set the face color of currently selected points and any added afterwards.'
            ),
        )
        self.borderColorEdit = QColorSwatchEdit(
            initial_color=self.layer.current_border_color,
            tooltip=trans._(
                'Click to set the border color of currently selected points and any added afterwards.'
            ),
        )
        self.faceColorEdit.color_changed.connect(self.changeCurrentFaceColor)
        self.borderColorEdit.color_changed.connect(
            self.changeCurrentBorderColor
        )

        sym_cb = QComboBox()
        sym_cb.setToolTip(
            trans._(
                'Change the symbol of currently selected points and any added afterwards.'
            )
        )
        current_index = 0
        for index, (symbol_string, text) in enumerate(
            SYMBOL_TRANSLATION.items()
        ):
            symbol_string = symbol_string.value
            sym_cb.addItem(text, symbol_string)

            if symbol_string == self.layer.current_symbol:
                current_index = index

        sym_cb.setCurrentIndex(current_index)
        sym_cb.currentTextChanged.connect(self.changeCurrentSymbol)
        self.symbolComboBox = sym_cb

        self.outOfSliceCheckBox = QCheckBox()
        self.outOfSliceCheckBox.setToolTip(trans._('Out of slice display'))
        self.outOfSliceCheckBox.setChecked(self.layer.out_of_slice_display)
        self.outOfSliceCheckBox.stateChanged.connect(self.change_out_of_slice)

        self.textDispCheckBox = QCheckBox()
        self.textDispCheckBox.setToolTip(trans._('Toggle text visibility'))
        self.textDispCheckBox.setChecked(self.layer.text.visible)
        self.textDispCheckBox.stateChanged.connect(self.change_text_visibility)

        self.select_button = self._radio_button(
            layer,
            'select_points',
            Mode.SELECT,
            True,
            'activate_points_select_mode',
        )
        self.addition_button = self._radio_button(
            layer,
            'add_points',
            Mode.ADD,
            True,
            'activate_points_add_mode',
        )

        self.delete_button = QtModePushButton(
            layer,
            'delete_shape',
        )
        action_manager.bind_button(
            'napari:delete_selected_points', self.delete_button
        )
        self._EDIT_BUTTONS += (self.delete_button,)
        self._on_editable_or_visible_change()

        self.button_grid.addWidget(self.delete_button, 0, 3)
        self.button_grid.addWidget(self.addition_button, 0, 4)
        self.button_grid.addWidget(self.select_button, 0, 5)

        self.layout().addRow(self.button_grid)
        self.layout().addRow(self.opacityLabel, self.opacitySlider)
        self.layout().addRow(trans._('blending:'), self.blendComboBox)
        self.layout().addRow(trans._('point size:'), self.sizeSlider)
        self.layout().addRow(trans._('symbol:'), self.symbolComboBox)
        self.layout().addRow(trans._('face color:'), self.faceColorEdit)
        self.layout().addRow(trans._('border color:'), self.borderColorEdit)
        self.layout().addRow(trans._('display text:'), self.textDispCheckBox)
        self.layout().addRow(trans._('out of slice:'), self.outOfSliceCheckBox)

    def changeCurrentSymbol(self, text):
        """Change marker symbol of the points on the layer model.

        Parameters
        ----------
        text : int
            Index of current marker symbol of points, eg: '+', '.', etc.
        """
        with self.layer.events.symbol.blocker(self._on_current_symbol_change):
            self.layer.current_symbol = SYMBOL_TRANSLATION_INVERTED[text]

    def changeCurrentSize(self, value):
        """Change size of points on the layer model.

        Parameters
        ----------
        value : float
            Size of points.
        """
        with self.layer.events.current_size.blocker(
            self._on_current_size_change
        ):
            self.layer.current_size = value

    def change_out_of_slice(self, state):
        """Toggleout of slice display of points layer.

        Parameters
        ----------
        state : QCheckBox
            Checkbox indicating whether to render out of slice.
        """
        # needs cast to bool for Qt6
        with self.layer.events.out_of_slice_display.blocker(
            self._on_out_of_slice_display_change
        ):
            self.layer.out_of_slice_display = bool(state)

    def change_text_visibility(self, state):
        """Toggle the visibility of the text.

        Parameters
        ----------
        state : QCheckBox
            Checkbox indicating if text is visible.
        """
        with self.layer.text.events.visible.blocker(
            self._on_text_visibility_change
        ):
            # needs cast to bool for Qt6
            self.layer.text.visible = bool(state)

    def _on_mode_change(self, event):
        """Update ticks in checkbox widgets when points layer mode is changed.

        Available modes for points layer are:
        * ADD
        * SELECT
        * PAN_ZOOM
        * TRANSFORM

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.

        Raises
        ------
        ValueError
            Raise error if event.mode is not ADD, PAN_ZOOM, TRANSFORM or SELECT.
        """
        super()._on_mode_change(event)

    def _on_text_visibility_change(self):
        """Receive layer model text visibiltiy change event and update checkbox."""
        with qt_signals_blocked(self.textDispCheckBox):
            self.textDispCheckBox.setChecked(self.layer.text.visible)

    def _on_out_of_slice_display_change(self):
        """Receive layer model out_of_slice_display change event and update checkbox."""
        with qt_signals_blocked(self.outOfSliceCheckBox):
            self.outOfSliceCheckBox.setChecked(self.layer.out_of_slice_display)

    def _on_current_symbol_change(self):
        """Receive marker symbol change event and update the dropdown menu."""
        with qt_signals_blocked(self.symbolComboBox):
            self.symbolComboBox.setCurrentIndex(
                self.symbolComboBox.findData(self.layer.current_symbol.value)
            )

    def _on_current_size_change(self):
        """Receive layer model size change event and update point size slider."""
        with qt_signals_blocked(self.sizeSlider):
            value = self.layer.current_size
            min_val = min(value) if isinstance(value, list) else value
            max_val = max(value) if isinstance(value, list) else value
            if min_val < self.sizeSlider.minimum():
                self.sizeSlider.setMinimum(max(1, int(min_val - 1)))
            if max_val > self.sizeSlider.maximum():
                self.sizeSlider.setMaximum(int(max_val + 1))
            with contextlib.suppress(TypeError):
                self.sizeSlider.setValue(int(value))

    @Slot(np.ndarray)
    def changeCurrentFaceColor(self, color: np.ndarray):
        """Update face color of layer model from color picker user input."""
        with self.layer.events.current_face_color.blocker(
            self._on_current_face_color_change
        ):
            self.layer.current_face_color = color

    @Slot(np.ndarray)
    def changeCurrentBorderColor(self, color: np.ndarray):
        """Update border color of layer model from color picker user input."""
        with self.layer.events.current_border_color.blocker(
            self._on_current_border_color_change
        ):
            self.layer.current_border_color = color

    def _on_current_face_color_change(self):
        """Receive layer.current_face_color() change event and update view."""
        with qt_signals_blocked(self.faceColorEdit):
            self.faceColorEdit.setColor(self.layer.current_face_color)

    def _on_current_border_color_change(self):
        """Receive layer.current_border_color() change event and update view."""
        with qt_signals_blocked(self.borderColorEdit):
            self.borderColorEdit.setColor(self.layer.current_border_color)

    def _on_ndisplay_changed(self):
        self.layer.editable = not (self.layer.ndim == 2 and self.ndisplay == 3)
        super()._on_ndisplay_changed()

    def close(self):
        """Disconnect events when widget is closing."""
        disconnect_events(self.layer.text.events, self)
        super().close()



# The tests in this module for the new style of async slicing in napari:
# https://napari.org/dev/naps/4-async-slicing.html
from functools import partial

import numpy as np
import pytest
from vispy.visuals import VolumeVisual

from napari import Viewer
from napari._tests.utils import LockableData
from napari._vispy.layers.base import VispyBaseLayer
from napari._vispy.layers.image import VispyImageLayer
from napari._vispy.layers.points import VispyPointsLayer
from napari._vispy.layers.vectors import (
    VispyVectorsLayer,
    generate_vector_meshes_2D,
)
from napari.layers import Image, Layer, Points, Vectors
from napari.utils.events import Event


@pytest.fixture
def rng() -> np.random.Generator:
    return np.random.default_rng(0)


@pytest.fixture
def _enable_async(_fresh_settings, make_napari_viewer):
    """
    This fixture depends on _fresh_settings and make_napari_viewer
    to enforce proper order of fixture execution.
    """
    from napari import settings

    settings.get_settings().experimental.async_ = True


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_image_on_current_step_change(
    make_napari_viewer, qtbot, rng
):
    viewer = make_napari_viewer()
    data = rng.random((3, 4, 5))
    image = Image(data)
    vispy_image = setup_viewer_for_async_slicing(viewer, image)
    assert viewer.dims.current_step != (2, 0, 0)

    viewer.dims.current_step = (2, 0, 0)

    wait_until_vispy_image_data_equal(qtbot, vispy_image, data[2, :, :])


@pytest.mark.usefixtures('_enable_async')
def test_async_out_of_bounds_layer_loaded(make_napari_viewer, qtbot):
    """Check that images that are out of bounds when slicing appear loaded.

    See https://github.com/napari/napari/issues/7070.
    """
    viewer = make_napari_viewer()
    l0 = viewer.add_image(np.random.random((5, 5, 5)))
    l1 = viewer.add_image(np.random.random((5, 5, 5)), translate=(5, 0, 0))
    assert viewer.dims.nsteps == (10, 5, 5)

    def layer_loaded(ly):
        return ly.loaded

    for i in range(viewer.dims.nsteps[0]):
        viewer.dims.current_step = (i, 0, 0)
        qtbot.waitUntil(partial(layer_loaded, l0), timeout=500)
        qtbot.waitUntil(partial(layer_loaded, l1), timeout=500)


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_image_on_order_change(make_napari_viewer, qtbot, rng):
    viewer = make_napari_viewer()
    data = rng.random((3, 5, 7))
    image = Image(data)
    vispy_image = setup_viewer_for_async_slicing(viewer, image)
    assert viewer.dims.order != (1, 0, 2)

    viewer.dims.order = (1, 0, 2)

    wait_until_vispy_image_data_equal(qtbot, vispy_image, data[:, 2, :])


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_image_on_ndisplay_change(make_napari_viewer, qtbot, rng):
    viewer = make_napari_viewer()
    data = rng.random((3, 4, 5))
    image = Image(data)
    vispy_image = setup_viewer_for_async_slicing(viewer, image)
    assert viewer.dims.ndisplay != 3

    viewer.dims.ndisplay = 3

    wait_until_vispy_image_data_equal(qtbot, vispy_image, data)


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_multiscale_image_on_pan(make_napari_viewer, qtbot, rng):
    viewer = make_napari_viewer()
    data = [rng.random((4, 8, 10)), rng.random((2, 4, 5))]
    image = Image(data)
    vispy_image = setup_viewer_for_async_slicing(viewer, image)

    # Force the image to be at the lower resolution, with #7870 this behavior
    # changed so that initial zoom was high, resulting in data_level=0.
    # Likely due to better sequencing of async slicing.
    viewer.camera.zoom = 0.1
    viewer.window._qt_viewer.canvas.on_draw(None)

    # Check that we're initially slicing the middle of the first dimension
    # over the whole of lowest resolution image.
    assert viewer.dims.not_displayed == (0,)
    assert viewer.dims.current_step[0] == 1
    assert image._data_level == 1
    np.testing.assert_equal(image.corner_pixels, [[0, 0, 0], [0, 3, 4]])

    # Simulate panning to the left by changing the corner pixels in the last
    # dimension, which corresponds to x/columns, then triggering a reload.
    image.corner_pixels = np.array([[0, 0, 0], [0, 3, 2]])
    image.events.reload(Event('reload', layer=image))

    wait_until_vispy_image_data_equal(qtbot, vispy_image, data[1][0, 0:4, 0:3])


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_multiscale_image_on_zoom(qtbot, make_napari_viewer, rng):
    viewer = make_napari_viewer()
    data = [rng.random((4, 8, 10)), rng.random((2, 4, 5))]
    image = Image(data)
    vispy_image = setup_viewer_for_async_slicing(viewer, image)

    # Force the image to be at the lower resolution, with #7870 this behavior
    # changed so that initial zoom was high, resulting in data_level=0.
    # Likely due to better sequencing of async slicing.
    viewer.camera.zoom = 0.1
    viewer.window._qt_viewer.canvas.on_draw(None)

    # Check that we're initially slicing the middle of the first dimension
    # over the whole of lowest resolution image.
    assert viewer.dims.not_displayed == (0,)
    assert viewer.dims.current_step[0] == 1
    assert image._data_level == 1
    np.testing.assert_equal(image.corner_pixels, [[0, 0, 0], [0, 3, 4]])

    # Simulate zooming into the middle of the higher resolution image.
    image._data_level = 0
    image.corner_pixels = np.array([[0, 2, 3], [0, 5, 6]])
    image.events.reload(Event('reload', layer=image))

    wait_until_vispy_image_data_equal(qtbot, vispy_image, data[0][1, 2:6, 3:7])


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_points_on_current_step_change(make_napari_viewer, qtbot):
    viewer = make_napari_viewer()
    data = np.array(
        [
            [0, 2, 3],
            [1, 3, 4],
            [2, 4, 5],
            [3, 5, 6],
            [4, 6, 7],
        ]
    )
    points = Points(data)
    vispy_points = setup_viewer_for_async_slicing(viewer, points)
    assert viewer.dims.current_step != (3, 0, 0)

    viewer.dims.current_step = (3, 0, 0)

    wait_until_vispy_points_data_equal(qtbot, vispy_points, np.array([[5, 6]]))


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_points_on_point_change(make_napari_viewer, qtbot):
    viewer = make_napari_viewer()
    # Define data so that slicing at 1.6 in the first dimension should match the
    # second point, but won't if that index is prematurely rounded as for other
    # layers.
    data = np.array(
        [
            [0, 2, 3],
            [1.4, 3, 4],
            [2.4, 4, 5],
            [3.4, 5, 6],
            [4, 6, 7],
        ]
    )
    points = Points(data)
    vispy_points = setup_viewer_for_async_slicing(viewer, points)
    assert viewer.dims.point != (1.6, 0, 0)

    viewer.dims.point = (1.6, 0, 0)

    wait_until_vispy_points_data_equal(qtbot, vispy_points, np.array([[3, 4]]))


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_image_loaded(make_napari_viewer, qtbot, rng):
    viewer = make_napari_viewer()
    data = rng.random((3, 4, 5))
    lockable_data = LockableData(data)
    layer = Image(lockable_data, multiscale=False)
    vispy_layer = setup_viewer_for_async_slicing(viewer, layer)

    assert layer.loaded
    assert viewer.dims.current_step != (2, 0, 0)

    with lockable_data.lock:
        viewer.dims.current_step = (2, 0, 0)
        assert not layer.loaded

    qtbot.waitUntil(lambda: layer.loaded)

    np.testing.assert_allclose(vispy_layer.node._data, data[2, :, :])


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_vectors_on_current_step_change(make_napari_viewer, qtbot):
    viewer = make_napari_viewer()
    data = np.array(
        [
            [[0, 2, 3], [1, 2, 2]],
            [[2, 4, 5], [0, -3, 3]],
            [[4, 6, 7], [3, 0, -2]],
        ]
    )
    vectors = Vectors(data)
    vispy_vectors = setup_viewer_for_async_slicing(viewer, vectors)
    assert viewer.dims.current_step != (2, 0, 0)

    viewer.dims.current_step = (2, 0, 0)

    wait_until_vispy_vectors_data_equal(
        qtbot, vispy_vectors, np.array([[[2, 4, 5], [0, -3, 3]]])
    )


@pytest.mark.usefixtures('_enable_async')
def test_async_slice_two_layers_shutdown(make_napari_viewer):
    """See https://github.com/napari/napari/issues/6685"""
    viewer = make_napari_viewer()
    # To reproduce the issue, we need two points layers where the second has
    # some non-zero coordinates.
    viewer.add_points()
    points = viewer.add_points()
    points.add([[1, 2]])

    viewer.close()


def setup_viewer_for_async_slicing(
    viewer: Viewer,
    layer: Layer,
) -> VispyBaseLayer:
    # Initially force synchronous slicing so any slicing caused
    # by adding the layer finishes before any other slicing starts.
    with viewer._layer_slicer.force_sync():
        # Add the layer and get the corresponding vispy layer.
        layer = viewer.add_layer(layer)
        vispy_layer = viewer.window._qt_viewer.layer_to_visual[layer]

    return vispy_layer


def wait_until_vispy_image_data_equal(
    qtbot, vispy_layer: VispyImageLayer, expected_data: np.ndarray
) -> None:
    def assert_vispy_image_data_equal() -> None:
        node = vispy_layer.node
        data = (
            node._last_data if isinstance(node, VolumeVisual) else node._data
        )
        # Vispy node data may have been post-processed (e.g. through a colormap),
        # so check that values are close rather than exactly equal.
        np.testing.assert_allclose(data, expected_data)

    qtbot.waitUntil(assert_vispy_image_data_equal)


def wait_until_vispy_points_data_equal(
    qtbot, vispy_layer: VispyPointsLayer, expected_data: np.ndarray
) -> None:
    def assert_vispy_points_data_equal() -> None:
        positions = vispy_layer.node._subvisuals[0]._data['a_position']
        # Flip the coordinates because vispy uses xy instead of rc ordering.
        # Also only take the number of dimensions expected since vispy points
        # are always 3D even when displaying 2D slices.
        data = positions[:, -expected_data.shape[1] :: -1]
        np.testing.assert_array_equal(data, expected_data)

    qtbot.waitUntil(assert_vispy_points_data_equal)


def wait_until_vispy_vectors_data_equal(
    qtbot, vispy_layer: VispyVectorsLayer, expected_data: np.ndarray
) -> None:
    def assert_vispy_vectors_data_equal() -> None:
        displayed = expected_data[..., -2:]
        exp_vertices, exp_faces = generate_vector_meshes_2D(
            displayed, 1, 1, 'triangle'
        )
        meshdata = vispy_layer.node._meshdata
        vertices = meshdata.get_vertices()
        faces = meshdata.get_faces()
        # invert for vispy
        np.testing.assert_array_equal(vertices, exp_vertices[..., ::-1])
        np.testing.assert_array_equal(faces, exp_faces)

    qtbot.waitUntil(assert_vispy_vectors_data_equal)
